{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name compose}
{@description Color composite}

{$I directives.pas}
uses simage,getparam;
var par         : tparams;
    cc          : CSImg_Config;
    ifn,fs,qs   : TAAStr;
    ofn,cfn,lfn : ansistring;
    qc,ql,fc,w  : double;
    f,q         : array of double;
    fi          : array of TSImg_FPImage;
    mi          : array of CSImg_MetaInfo;
    fj          : TSImg_FPImage;
    i,j,k,c     : longint;
    d           : ^simgfp;
    da          : array of ^simgfp;
    r,s,t       : double;
    crop        : boolean;

begin
    with par do
    begin
        init;
        paramusage:='<infiles>';
        ifn.size:=0;    defparam('<infiles>','Input images (can be specified multiple times)',ifn);
        cfn:='';        defparam('-c','Add this continuum image to all colors. This is done before LRGB transformation',cfn);   deffmt('-c','<infile>');
        lfn:='';        defparam('-l','LRGB-like composition using this luminance image',lfn);  deffmt('-l','<infile>');
        fs.size:=0;     defparam('-f','Factors (can be specified multiple times, default: 1)',fs);
        fc:=1;          defparam('-fc','Factor for continuum image (default: 1)',fc);
        qs.size:=0;     defparam('-q','Zero level quantiles (can be specified multiple times, <0 to ignore, default: -1)',qs);
        ql:=-1;         defparam('-ql','Zero level quantile for luminance image. Negative values will be cropped.',ql);
        qc:=-1;         defparam('-qc','Zero level quantile for continuum image. Negative values will be cropped.',qc);
        w:=2.5;         defparam('-w','Desired FWHM for low pass filtering colors in LRGB-like mode. default 2.5',w);
        crop:=false;    defparam('-n','Crop negative values.',crop);
        ofn:='';        defparam('-o','Output image (default: stdout)',ofn);            deffmt('-o','<outfile>');
                        defparam('-h','help',helpf);
        run;
        done;
    end;
    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug('compose: ifn.size='+int2str(ifn.size)+' fs.size='+int2str(fs.size)+' qs.size='+int2str(qs.size)+' cfn='+cfn+' lfn='+lfn+' ofn='+ofn+' w='+real2str(w)+' crop='+boolean2str(crop));

    { validate parameters }
    if ifn.size<1 then cc.error('compose: At least two input images required');
    setlength(f,ifn.size);
    setlength(q,ifn.size);
    setlength(fi,ifn.size);
    setlength(mi,ifn.size);
    setlength(da,ifn.size);
    for i:=0 to ifn.size-1 do
    begin
        fi[i].pt:=nil;
        if i>=fs.size then f[i]:=1
        else if not val2(fs.aa[i],f[i]) then cc.error('add: Number expected after -f, got '+fs.aa[i]);
        if i>=qs.size then q[i]:=-1
        else if not val2(qs.aa[i],q[i]) then cc.error('add: Number expected after -q, got '+qs.aa[i]);
    end;

    { read inputs }
    for j:=0 to ifn.size-1 do
    begin
        mi[j]:=CSImg_MetaInfo.create;
        cc.Info('compose: Reading input image '+ifn.aa[j]);
        SImg_ReadImageFile(cc, ifn.aa[j], fi[j], mi[j]);
        if (fi[j].nx<>fi[0].nx) or (fi[j].ny<>fi[0].ny) then cc.error('compose: Invalid dimensions of '+ifn.aa[j]);

        if (q[j]>-simg_mina) or (abs(f[j]-1)>1e-10) then
        begin
            if q[j]>-simg_mina then r:=SImg_GetQuantile(cc, fi[j], q[j])
            else r:=0;
            s:=f[j];
            d:=fi[j].pt;
            if crop then
                for i:=0 to fi[j].size-1 do
                begin
                    t:=(d[i]-r)*s;
                    if t<0 then t:=0;
                    d[i]:=t;
                end
            else
                for i:=0 to fi[j].size-1 do
                    d[i]:=(d[i]-r)*s;
        end;
    end;

    { continuum image }
    if cfn<>'' then
    begin
        fj.pt:=nil;
        cc.Info('compose: Reading continuum image '+cfn);
        SImg_ReadImageFile(cc,cfn,fj);
        if (fj.nc<>1) or (fj.nx<>fi[0].nx) or (fj.ny<>fi[0].ny) then cc.error('compose: Invalid dimensions of continuum image');
        if qc>=-simg_mina then qc:=SImg_GetQuantile(cc, fj, qc)
        else qc:=0;
        k:=ifn.size-1;
        for j:=0 to ifn.size-1 do
            da[j]:=fi[j].pt-1;
        for i:=0 to fj.size-1 do
        begin
            r:=(fj.pt[i]-qc)*fc;
            if r<=0 then
                for j:=0 to k do
                    da[j]+=fi[j].nc
            else
                for j:=0 to k do
                begin
                    for c:=1 to fi[j].nc do
                        da[j][c]+=r;
                    da[j]+=fi[j].nc;
                end;
        end;
        SImg_Free(cc,fj);
    end;

    { color composition }
    fj.pt:=nil;
    if lfn<>'' then
    begin
        for i:=0 to ifn.size-1 do
        begin
            if mi[i].get('FWHMPIX',f[i]) then cc.Info('compose: Read PSF FWHM metadata of input image '+int2str(i+1)+': ' +real2str(f[i],-1))
            else begin
                f[i]:=SImg_GetSharpnessFWHM(cc,fi[i]);
                cc.Info('compose: Auto-detected PSF FWHM width from input image '+int2str(i+1)+': ' +real2str(f[i],-1));
            end;
        end;
        for i:=0 to ifn.size-1 do
            if ifn.aa[i]=lfn then       { re-use file }
            begin
                if (ql>-simg_mina) and (abs(ql-q[i])>1e-5) then SImg_Duplicate(cc, fi[i],fj, true)
                else begin
                    fj:=fi[i];
                    ql:=-1;
                end;
                break;
            end;
        if fj.pt=nil then
        begin
            cc.Info('compose: Reading luminance image '+lfn);
            SImg_ReadImageFile(cc,lfn,fj);
        end;
        if (fj.nc<>1) or (fi[0].nx<>fj.nx) or (fi[0].ny<>fj.ny) then cc.error('compose: Invalid dimensions of luminance image');
        if ql>=-simg_mina then
        begin
            r:=SImg_GetQuantile(cc, fj, ql);
            for i:=0 to fj.size-1 do
                fj.pt[i]-=r;
        end;
        cc.Info('compose: Color composition with luminance image');
        SImg_LCompose(cc, fi,fj, fj , f, w);
    end
    else begin
        cc.Info('compose: Color composition');
        SImg_Compose(cc, fi,fj);
    end;
    for i:=0 to ifn.size-1 do
        SImg_Free(cc,fi[i]);

    { output }
    mi[0].deleteRaw;
    mi[0].delete('NOISEFLR');
    cc.Info('compose: Writing output image '+ofn);
    SImg_WriteImageFile(cc,ofn,fj,mi[0]);
    SImg_Free(cc,fj);
    for j:=0 to ifn.size-1 do
        mi[j].destroy;

    cc.destroy;
end.
