{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name denoise2}
{@description Denoising by low-pass filtering pixels within constraints.}

{$I directives.pas}
uses simage,getparam;
var par                 : tparams;
    cc                  : CSImg_Config;
    fid,fic             : TSImg_FPImage;
    rm                  : TSImg_RDImage;
    fidn,ficn,fimn,fon  : ansistring;
    thr,dn,dnn,dnc,rmax : double;
    fwhm,sig            : double;
    pm                  : longint;
    mic,mid             : CSImg_MetaInfo;

begin
    with par do
    begin
        init;
        paramusage:='[<infile> [<outfile>]]';
        fidn:='';    defparam('<infile>','Input image (default: stdin)',fidn);
        fimn:='';    defparam('-m','Input interpolation mask. Can be specified multiple times',fimn);    deffmt('-m','<infile>');
        fon:='';     defparam('<outfile>','Output image (default: stdout)',fon);
        ficn:='';    defparam('-ic','Input image, used for determination of constraints. Normal input image is used by default.',ficn);
        thr:=0.95;   defparam('-t','Denoising threshold: no denoising above that quantile. Default: 95%. Can be increased if image contains no stars.', thr);
        dn:=1.5;     defparam('-n','Median upper constraint relative to noise in center. Default: 1.5', dn);
        dnn:=2;      defparam('-nn','Lower constraint relative to upper constraint. Default: 2', dnn);
        dnc:=1;      defparam('-nc','Constraint in corner relative to constraint in center. Default: 1', dnc);
        rmax:=250;   defparam('-r','Maximum filter radius: Default: 250',rmax);
        fwhm:=-1;    defparam('-w','Input FWHM. Default: auto-detect.',fwhm);
        sig:=-1;     defparam('-s','Input noise level / detection level. 0: auto-detect. Default: Try to read from meta info or auto-detect.',sig);
        pm:=1;       defparam('-p','Pattern for interpolation mask. Interpolation if mask and pattern is nonzero. Default: 1',pm);
                     defparam('-o','Destination image',fon); deffmt('-o','<outfile>');
                     defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc:=CSImg_Config.create(SImg_Config_Default);
    cc.debug(cc.name+': fidn='+fidn+' ficn='+ficn+' fimn='+fimn+' fon='+fon+' thr='+real2str(thr)+' dn='+real2str(dn)+' dnn='+real2str(dnn)+' dnc='+real2str(dnc)+' rmax='+real2str(rmax)+' fwhm='+real2str(fwhm)+' sig='+real2str(sig)+' pm='+int2str(pm));

    fid.pt:=nil;
    mid:=CSImg_MetaInfo.create;
    cc.Info(cc.name+': Reading input image '+fidn);
    SImg_ReadImageFile(cc,fidn,fid,mid);

    if ficn='' then
    begin
        mic:=mid;
        fic:=fid;
    end
    else begin
        fic.pt:=nil;
        mic:=CSImg_MetaInfo.create;
        cc.Info(cc.name+': Reading input image '+ficn);
        SImg_ReadImageFile(cc,ficn,fic,mic);
        if (fic.nx<>fid.nx) or (fic.ny<>fid.ny) or (fic.nc<>fid.nc) then cc.error(cc.name+': Incompatible dimensions of input images');
    end;

    rm.pt:=nil;
    if fimn<>'' then
    begin
        cc.Info(cc.name+': Reading interpolation mask');
        SImg_ReadImageFile(cc, fimn, rm);
        if rm.fmt<>SImg_RawByte then cc.error(cc.name+': Invalid format interpolation mask '+fimn);
        if (rm.nx<>fid.nx) or (rm.ny<>fid.ny) or (rm.nc<>fid.nc) then cc.error(cc.name+': Invalid dimensions of interpolation mask '+fimn);
    end;

    if fwhm<0.1 then
    begin
        if mic.get('FWHMPIX',fwhm) then cc.Info(cc.name+': Read PSF FWHM width from input image: ' +real2str(fwhm,-1))
        else begin
            fwhm:=SImg_GetSharpnessFWHM(cc,fic);
            cc.Info(cc.name+': Auto-detected input FWHM: ' +real2str(fwhm,-1));
        end;
    end;

    if sig<simg_mina then
    begin
        if (sig<-simg_mina) and mic.get('NOISEFLR',sig) then cc.Info(cc.name+': Read noise floor from input image: ' +real2str(sig))
        else begin
            sig:=SImg_NoiseFloor(cc, fic, 1+fwhm*0.5);
            cc.Info(cc.name+': Autodetected noise floor: ' +real2str(sig));
        end;
    end;

    cc.Info(cc.name+': Denoising using up to ' +int2str(cc.threads) + ' threads');
    SImg_Denoise2(cc, fid,fic, rm, fid, thr, sig*dn,dnn,dnc, 0.4*fwhm,rmax, pm);
    SImg_Free(cc,fic);
    SImg_Free(cc,rm);

    cc.Info(cc.name+': Writing output file '+fon);
    mid.add('NOISEFLR', real2str(sig,10),'Noise floor / detection limit');
    SImg_WriteImageFile(cc,fon,fid,mid);
    SImg_Free(cc,fid);


    cc.destroy;
end.
