{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name sdbconvert}
{@description Converts stack database to image.}

{ Option -l dims bright stars }

{$I directives.pas}
uses simage,getparam,sysutils;
var par                 : tparams;
    ifn,vfn,wfn,sfn     : ansistring;
    ci,snr,bg           : double;
    sm,sc,si            : boolean;

    cc                  : CSImg_Config;
    fs,fi,fw            : TSImg_FPImage;
    mi                  : CSImg_MetaInfo;
    x1,y1,x2,y2,i       : longint;
    r                   : double;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);
    with par do
    begin
        init;
        paramusage:='[<infile>]';
        ifn:='';    defparam('<infile>','Input stack data base',ifn);
        vfn:='';    defparam('-v','Write values to this file',vfn);                     deffmt('-v','<outfile>');
        sfn:='';    defparam('-s','Write SNR to this file',sfn);                        deffmt('-s','<outfile>');
        wfn:='';    defparam('-w','Write weight to this file',wfn);                     deffmt('-w','<outfile>');
        sc:=false;  defparam('-sc','Separate color mode.',sc);
        snr:=-1;    defparam('-l','Resolution optimization mode by limiting SNR.\\Interpreted as quantile if <=1, otherwise as SNR.\\This may dim bright stars.' ,snr);
        bg:=-1;     defparam('-b','Constant background subtraction. Parameter is the median filter radius.' ,bg);
        ci:=-1;     defparam('-ic','Print region crop region info. Parameter is a threshold typically between 0.2 and 0.8',ci);
        si:=false;  defparam('-is','Print sharpness info.',si);
        sm:=false;  defparam('-sm','Script friendly crop information',sm);
                    defparam('-h','Help',helpf);
        run;
        done;
    end;

    cc.debug(cc.name+': in='+ifn+' out_vals='+vfn+' out_weights='+wfn+'  out_snr='+sfn+'  ci='+real2str(ci)+'  si='+boolean2str(si)+'  sc='+boolean2str(sc)+'  snr='+real2str(snr)+'  bg='+real2str(bg));

    if (vfn='') and (wfn='') and (sfn='') and (sfn='') and (ci<=0) then
    begin
        cc.Info(cc.name+': Nothing to do');
        cc.destroy();
        SImg_Halt(0);
    end;

    if (snr>0) and (sc) then cc.error(cc.name+': -sc and -l cant be defined at the same time');

    cc.Info(cc.name+': Reading stack database '+ifn);
    fs.pt:=nil;
    mi:=CSImg_MetaInfo.create();
    SImg_ReadImageFile(cc,ifn,fs,mi);
    SImg_StackDB_correct(cc,fs);

    if ci>0 then
    begin
        SImg_StackDB_CropInfo(cc, fs, ci, x1,y1,x2,y2);
        if sm then writeln(x1:5, ' ', y1:5,'    ',x2:5 , ' ', y2:5)
        else writeln('Calculated crop region: (', x1:5 , ',', y1:5,') ... (',x2:5 , ',', y2:5,')');
    end;

    fi.pt:=nil;  fw.pt:=nil;
    if snr>0 then cc.info(cc.name+': Resolution optimized conversion using '+int2str(cc.threads)+' threads')
    else cc.info(cc.name+': Conversion');
    SImg_StackDB_convert(cc,fs,fi,fw,snr);
    SImg_Free(cc,fs);

    if (bg>0.1) and ((vfn<>'') or (sfn<>'')) then
    begin
        cc.info(cc.name+': Estimating background using '+int2str(cc.threads)+' threads');
        bg:=SImg_BgApprox5(cc, fi,fw, 0,bg);
        for i:=0 to fi.size-1 do
            fi.pt[i]-=bg;
    end;

    if si then
    begin
        r:=SImg_GetSharpnessFWHM(cc, fi);
        cc.info(cc.name+': New FWHM='+real2str(r,-1));
        mi.add('FWHMPIX',real2str(r,5),'Star FWHM in pixels');
    end;

    if vfn<>'' then
    begin
        cc.Info(cc.name+': Writing values to image '+vfn);
        SImg_WriteImageFile(cc,vfn,fi,mi);
    end;

    if wfn<>'' then
    begin
        cc.Info(cc.name+': Writing weights to image '+wfn);
        SImg_WriteImageFile(cc,wfn,fw,mi);
    end;

    if sfn<>'' then
    begin
        cc.Info(cc.name+': Writing SNR to image '+sfn);
        for i:=0 to fi.size-1 do
        fi.pt[i]*=sqrt(fw.pt[i]);
        SImg_WriteImageFile(cc,sfn,fi,mi);
    end;
    SImg_Free(cc,fw);
    SImg_Free(cc,fi);

    cc.destroy;
end.
