{  simg -- command line applications for deterministic image processing
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{@name starconv}
{@description Transforms star image and optionally add it to an input image.}

{$I directives.pas}
uses simage,getparam,sysutils;

const hist_n = 65536;

var par         : tparams;
    cc          : CSImg_Config;
    ifn,sfn,ofn : ansistring;
    tfn,mfn,osfn: ansistring;
    thr,ofwhm   : double;
    mn,mm       : longint;

    hist        : array[0..hist_n] of longint;
    fi,fs,ft    : TSImg_FPImage;
    mi          : CSImg_MetaInfo;
    inf         : TSImg_PSFInfo;
    rm          : TSImg_RDImage;
    i,x,y       : longint;
    r,s         : double;
    a,b         : simgfp;
    d           : ^simgfp;
    e           : ^byte;

begin
    cc:=CSImg_Config.create(SImg_Config_Default);

    with par do
    begin
        init;
        paramusage:='<infile1> [<infile2> [<outfile>]]';
        ifn:='';        defparam('<infile1>','Deconvoluted star image',@ifn,tpastr,rapdreq);
        sfn:='';        defparam('<infile2>','Input PSF. Default: stdin',sfn);
        ofn:='';        defparam('<outfile>','Output image (default: stdout)', ofn);
        thr:=1;         defparam('-t','Reduce number of star to that fraction. Default: 1', thr );
        tfn:='';        defparam('-ti','Use this image instead of input image for determination stars to ignore',tfn);   deffmt('-ti','<intfile>');
        mfn:='';        defparam('-mi','Interpolation mask. Ignore stars if less than n (see -mn) of the 9 surrounding pixels pixels are interpolated',mfn);     deffmt('-mi','<intfile>');
        mn:=6;          defparam('-mn','Ignore threshold in combination with -mi. default: 6',mn);
        mm:=1;          defparam('-mm','Interpolation mask m. If mask and m<>0, pixel is considered as interpolated. default: 1',mn);
        ofwhm:=-1;      defparam('-f','(De)convolute PSF to that FWHM.',ofwhm);
                        defparam('-o','Output image',ofn);                         deffmt('-o','<outfile>');
        osfn:='';       defparam('-os','Output PSF (useful in combination with -f)',osfn);      deffmt('-os','<outfile>');
                        defparam('-h','Help',helpf);
        run;
        done;
    end;
    cc.debug(cc.name+': ifn='+ifn+' sfn='+sfn+' tfn='+tfn+' mfn='+mfn+' ofn='+ofn+' osfn='+osfn+' thr='+real2str(thr)+' mn='+int2str(mn)+' mm='+int2str(mm)+' ofwhm='+real2str(ofwhm));

    { input image }
    mi:=CSImg_MetaInfo.create();
    cc.info(cc.name+': Reading input image '+ifn);
    SImg_ReadImageFile(cc, ifn, fi, mi);
    if fi.nc<>1 then cc.info(cc.name+': Only 1-color images supported');

    { input PSF }
    cc.info(cc.name+': Reading PSF '+sfn);
    SImg_ReadImageFile(cc, sfn, fs);
    SImg_GetPSFInfo(cc,fs,inf);

    { (de)convolute PSF }
    if ofwhm>0.1 then SImg_DeconvPSF(cc, fs,fs, ofwhm);
    if osfn<>'' then SImg_WriteImageFile(cc, osfn, fs);

    { threshold }
    if thr<0 then thr:=0;
    thr:=1-thr;
    if thr>1e-3 then
    begin
        ft.pt:=nil;
        if tfn='' then ft:=fi
        else begin
            cc.info(cc.name+': Reading threshold reference image '+tfn);
            SImg_ReadImageFile(cc, tfn, ft);
            if ft.size<>fi.size then cc.error(cc.name+': Incompatible dimension of input given by -ti');
        end;
        r:=0; s:=0;
        for i:=0 to ft.size-1 do
        begin
            a:=ft.pt[i];
            if a<=0 then continue;
            r+=sqr(a);
            s+=a;
        end;
        if r<0 then cc.error(cc.name+': no stars found');
        s:=s*sqr(hist_n)/r;
        fillchar(hist,sizeof(hist),0);
        for i:=0 to ft.size-1 do
        begin
            a:=ft.pt[i];
            if a<simg_mina then continue;
            x:=round(sqrt(a*s));
            if x>hist_n then x:=hist_n;
            hist[x]+=1;
        end;
        x:=0;
        for i:=0 to hist_n do
            x+=hist[i];
        x:=round(x*thr-0.49);
        i:=0; y:=0;
        while y<x do
        begin
            y+=hist[i];
            i+=1;
        end;
        a:=sqr(i)/s;
        for i:=0 to ft.size-1 do
        begin
            b:=ft.pt[i];
            if b<=a then fi.pt[i]:=0
            else fi.pt[i]*=sqr((b-a)/b);
        end;
        if ft.pt<>fi.pt then SImg_Free(cc,ft);
    end;

    { interpolation mask }
    if (mfn<>'') and (mn>0) then
    begin
        if mn>9 then mn:=9;
        cc.info(cc.name+': Reading interpolation mask '+mfn);
        SImg_ReadImageFile(cc, mfn, rm);
        if rm.fmt<>SImg_RawByte then cc.error(cc.name+': 8 bit image expected');
        if rm.size<>fi.size then cc.error(cc.name+': Incompatible dimension of input given by -mi');
        if rm.nc<>1 then cc.error(cc.name+': Incompatible dimension of input given by -mi');
        for y:=1 to fi.ny-2 do
        begin
            e:=rm.ptb + rm.sl*y + 1;
            d:=fi.pt  + fi.sl*y + 1;
            for x:=0 to fi.nx-2 do
            begin
                if d^<simg_mina then d^:=0
                else begin
                    i:=0;
                    if e[-rm.sl-1] and mm<>0 then i+=1;
                    if e[-rm.sl  ] and mm<>0 then i+=1;
                    if e[-rm.sl+1] and mm<>0 then i+=1;
                    if e[      -1] and mm<>0 then i+=1;
                    if e[       0] and mm<>0 then i+=1;
                    if e[       1] and mm<>0 then i+=1;
                    if e[ rm.sl-1] and mm<>0 then i+=1;
                    if e[ rm.sl  ] and mm<>0 then i+=1;
                    if e[ rm.sl+1] and mm<>0 then i+=1;
                    if i<mn then d^:=0;
                end;
                d+=1; e+=1;
            end;
        end;
        SImg_Free(cc, rm);
    end;

    { interpolation mask }
    cc.info(cc.name+': Convolution');
    SImg_Convolute(cc, fi,fs,fi, round(inf.xc),round(inf.yc));
    SImg_Free(cc, fs);

    { output image }
    cc.Info(cc.name+': Writing output image '+ofn);
    SImg_WriteImageFile(cc, ofn, fi, mi);
    SImg_Free(cc,fi);

    mi.destroy;
    cc.destroy;
end.
