{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{$PACKRECORDS 8}
{$PACKENUM 4}     { important: must be 4 }

const SImg_SIFHeaderID='zFiS';
      SImg_SIFHeader_size=128;

type TSImg_SIFHeader=packed record
       case byte of
          0 : ( hid           : array[0..3] of char; { must be SImg_SIFHeaderID }
{4}             first         : longint;             { absolute address of first block; must be >=128 }
{8}             xsize,ysize   : longint;             { dimension }
{16}            cn            : longint;             { number of colors }
{20}            data          : TSImg_RawFormat;     { any coding supported }
{24}            minval,maxval : double;              { value of 0 and 255/65535/1 }
{40}            xc,yc         : longint;             { center }
{48}            maxa          : double;              { maximal valid absolute value }
              );
          1 : ( dummy: array[0..SImg_SIFHeader_size-1] of byte
              );
       end;

const SImg_SIFBlockID='YfIs';

type TSImg_SIFBlockID=(SImg_SIFBlock_BTM:=0, SImg_SIFBlock_Dummy, SImg_SIFBlock_Ext_Exposure, SImg_SIFBlock_Ext_Bayer);

type TSImg_SIFBlock=packed record
{0}     bid          : array[0..3] of char;  { must be SIFBlockID }
{4}     next         : longint;              { next block relative to start of data; <0 if last block }
{8}     id           : TSImg_SIFBlockID;     { SIFBlockBTM or SIFBlockExtension }
{12}    length       : longint;              { data size (normally length=next) }
        end;

type TSImg_SIFExt_Exposure=packed record
       adr       : double;    { AD resolution in ADU/e, (<AD units>/<electrons>) }
       sron      : double;    { square read out noise }
       ofl,ufl   : double;    { overflow/underflow limit }
       ofld,ufld : boolean;   { overflow/underflow defined }
       hbl,vbl   : boolean;   { horizontal / vertical blooming }
       ne        : longint;   { number of (different time) exposures (fi.nc mod ni = 0 !) }
{       t        : array [0..ne-1] of double;  integration times  }
       end;

{ possible patterns:
  0: r g    1: g r    2: g b    3: b g
     g b       b g       r g       g r }
type TSImg_SIFExt_Bayer=packed record
       pattern   : longint;
       end;

@DEFAULT_PACKRECORDS;

{*************************************************************************}
{*************************************************************************}
{******* SIF (image format) routines *************************************}
{*************************************************************************}
{*************************************************************************}

{*************************************************************************}
{******* SImg_RSIF *******************************************************}
{*************************************************************************}
function SImg_RSIF(cc:CSImg_Config; fb:TSImg_FileBuf; var ri:TSImg_RDImage; mi:CSImg_MetaInfo):boolean;
var sh       : ^TSImg_SIFHeader;
    sb       : ^TSImg_SIFBlock;
    d        : qword;
    lb       : boolean;
    expb     : ^TSImg_SIFExt_Exposure;
    bpb      : ^TSImg_SIFExt_Bayer;
    pd       : ^double;
    i        : longint;
    s        : string7;

begin
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_RSIF: called with: ri='+SImg_InfoStr2(ri)+' '+int2str(fb.size)+' Bytes input buffer');
    SImg_Free(cc,ri);
    SImg_RSIF:=false;
    sh:=@fb.pt^;
    if (sh^.hid<>SImg_SIFHeaderID) or (sh^.first<56) then exit;

    try
        with sh^ do
        begin
            SImg_New(cc, ri, data, xsize, ysize, cn, false);
            ri.co:=minval;
            ri.cf:=maxval-minval;
            if data=SImg_RawByte then ri.cf/=255
            else if data=SImg_RawWord then ri.cf/=65535;
        end;
    except
        exit;
    end;

    expb:=nil;
    bpb:=nil;
    d:=sh^.first;
    lb:=false;
    while (d+sizeof(TSImg_SIFBlock)<=fb.size) and (not lb) do
    begin
        sb:=@fb.pt[d];
        d+=sizeof(TSImg_SIFBlock);
        if sb^.bid<>SImg_SIFBlockID then
        begin
            cc.warning('SImg_RSIF: Invalid block ID');
            exit;
        end;
        if (d+sb^.length>fb.size) or ((sb^.length>sb^.next) and (sb^.next>=0)) then
        begin
            cc.warning('SImg_RSIF: Invalid block size');
            exit;
        end;
        case sb^.id of
            SImg_SIFBlock_BTM :
                begin
                    if SImg_RSIF then cc.warning('SImg_RSIF: Only one btm block allowed')
                    else if (sb^.length<>ri.dsize) then cc.warning('SImg_RSIF: Invalid size of btm block')
                    else begin
                        move(fb.pt[d], ri.pt^, sb^.length);
                        SImg_RSIF:=true;
                    end;
                end;
            SImg_SIFBlock_Dummy :
                begin
                end;
            SImg_SIFBlock_Ext_Exposure :
                begin
                    expb:=@fb.pt[d];
                end;
            SImg_SIFBlock_Ext_Bayer :
                begin
                    bpb:=@fb.pt[d];
                end;
            else cc.warning('SImg_RSIF: Unknown block type');
        end;
        if sb^.next<0 then lb:=true
        else d+=sb^.next;
    end;

    if SImg_RSIF then
    begin
        if (expb<>nil) and( mi<>nil) then
        begin
            mi.add('BUNIT','''adu''','Image unit');
            pd:=@expb[1];
            mi.add('EXPTIME',real2str(pd^,8),'Exposure time in s of frame 1');
            for i:=1 to expb^.ne-1 do
                mi.add('ETIME_'+int2str(i),real2str(pd[i],8),'Exposure time in s of frame '+int2str(i+1));
            if expb^.adr>1e-6 then mi.add('EGAIN',real2str(1/expb^.adr,8),'ADC resolution in electrons / ADU');
            if expb^.sron>0 then mi.add('RDNOISE',real2str(sqrt(expb^.sron),8),'Read out noise in ADU');
            if expb^.ufld then mi.add('DATAMIN',real2str(expb^.ufl,15),'Minimum valid value');
            if expb^.ofld then mi.add('DATAMAX',real2str(expb^.ofl,15),'Maximum valid value');
        end;
        if (bpb<>nil) and (mi<>nil) then
        begin
            s:='';
            case bpb^.pattern of
                0 : s:='''RG/GB''';
                1 : s:='''GR/BG''';
                2 : s:='''GB/RG''';
                3 : s:='''BG/GR''';
            end;
            if s<>'' then mi.add('BAYERPAT',s,'Bayer color pattern');
        end;
    end
    else cc.warning('SImg_RSIF: No btm block found');
    if not lb then cc.warning('SImg_RSIF: Unexpected end of file');
end;


{*************************************************************************}
{******* SImg_WSIF *******************************************************}
{*************************************************************************}
procedure SImg_WSIF(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo);
var sh     : TSImg_SIFHeader;
    sb     : TSImg_SIFBlock;
    expb   : TSImg_SIFExt_Exposure;
    etime  : array[0..99] of double;
    bpb    : TSImg_SIFExt_Bayer;
    s      : ansistring;
    bps    : string7;
    cf,r   : double;
    b      : boolean;
    i      : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_WSIF: called with: output='+f.info+' ri='+SImg_InfoStr2(ri));
    if not SImg_Recover(cc,ri) then SImg_Error('SImg_WSIF: Invalid image');

    if not ri.bole then SImg_SwapBytes(ri);

    fillchar(sh,sizeof(sh),0);
    sh.data:=ri.fmt;
    sh.hid:=SImg_SIFHeaderID;
    sh.first:=sizeof(sh);
    sh.xsize:=ri.nx; sh.ysize:=ri.ny; sh.cn:=ri.nc;
    sh.maxa:=1e200;
    cf:=ri.cf;
    if sh.data=SImg_RawByte then cf*=255
    else if sh.data=SImg_RawWord then cf*=65535;
    sh.minval:=ri.co;
    sh.maxval:=ri.co+cf;

    f.open(cc);
    f.write(cc,@sh,sizeof(sh));
    sb.bid:=SImg_SIFBlockID;
    sb.next:=ri.dsize;
    sb.id:=SImg_SIFBlock_BTM;
    sb.length:=ri.dsize;
    f.write(cc,@sb,sizeof(sb));
    f.write(cc,ri.pt,sb.length);

    if mi<>nil then
    begin
        // exposure information
        b:=false;
        expb.adr:=0;
        if mi.get('EGAIN',r) and (r>1e-8) then
        begin
            expb.adr:=1/r;
            b:=true;
        end;
        expb.sron:=1;
        if mi.get('RDNOISE',r) then
        begin
            expb.sron:=sqr(r);
            b:=true;
        end;
        expb.ofl:=1e200;
        expb.ofld:=mi.get('DATAMAX',expb.ofl);
        expb.ufl:=-1e200;
        expb.ufld:=mi.get('DATAMIN',expb.ufl);
        expb.hbl:=false;
        expb.vbl:=false;
        i:=0;
        if mi.get('EXPTIME',r) then
        begin
            etime[0]:=r;
            repeat
                inc(i);
            until (i>99) or (not mi.get('ETIME_'+int2str(i),etime[i]));
        end;
        expb.ne:=i;
        if b or expb.ofld or expb.ufld or (i>0) then
        begin
            sb.bid:=SImg_SIFBlockID;
            sb.next:=expb.ne*sizeof(double)+sizeof(expb);
            sb.id:=SImg_SIFBlock_Ext_Exposure;
            sb.length:=sb.next;
            f.write(cc,@sb,sizeof(sb));
            f.write(cc,@expb,sizeof(expb));
            f.write(cc,@etime,i*sizeof(double));
        end;

        // bayer information
        if mi.get('BAYERPAT',s) then
        begin
            bps:=trimquotes(s);
            if length(bps)=4 then insert('/',bps,3);
            i:=-1;
            if bps='RG/GB' then i:=0
            else if bps='GR/BG' then i:=1
            else if bps='GB/RG' then i:=2
            else if bps='BG/GR' then i:=3
            else if bps='12/23' then i:=0
            else if bps='21/22' then i:=1
            else if bps='22/12' then i:=2
            else if bps='22/21' then i:=3;
            if i<0 then cc.warning('SImg_WSIF: invalid Bayer pattern'+bps);
            bpb.pattern:=i;
            sb.bid:=SImg_SIFBlockID;
            sb.next:=sizeof(bpb);
            sb.id:=SImg_SIFBlock_Ext_Bayer;
            sb.length:=sb.next;
            f.write(cc,@sb,sizeof(sb));
            f.write(cc,@bpb,sizeof(bpb));
        end;
    end;

    sb.bid:=SImg_SIFBlockID;
    sb.next:=-1;
    sb.id:=SImg_SIFBlock_Dummy;
    sb.length:=0;
    f.write(cc,@sb,sizeof(sb));
    f.close(cc);
end;
