{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

const SImg_TSIFHeader_size=128; // with ID

type TSImg_TSIFHeader=packed record             { without id }
       case byte of
          0 : ( nx,ny,nc      : longint;        { 4, image dimensions }
                dsize         : int64;          { 16, data size in bytes }
                cf,co         : double;         { 24, conversion factor and offset }
                fmt0          : byte;           { 40, ri.fmt, any coding supported }
                fmt1          : byte;           { 41, bit 0: byte order is little endian }
                ms            : word;           { 42..43, number of meta information }
              );
          1 : ( dummy: array[0..SImg_SIFHeader_size-5] of byte;
              );
       end;

type TSImg_TSIFMeta=packed record
        ks,vs  : byte;  { key size, value size }
        b      : array[0..125] of char;  { key, value, comment (null terminated) }
        end;

{*************************************************************************}
{******* SImg_RTSIF *******************************************************}
{*************************************************************************}
procedure SImg_RTSIF(cc:CSImg_Config; fd:longint; var ri:TSImg_RDImage; mi:CSImg_MetaInfo; headerOnly:boolean);
var header              : TSImg_TSIFHeader;
    md                  : ^TSImg_TSIFMeta;
    key,val,comment     : ansistring;
    i,cs                : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_RTSIF: called with: ri='+SImg_InfoStr2(ri)+' headerOnly='+boolean2str(headerOnly));
    if fpread2(fd, @header, sizeof(header))<sizeof(header) then cc.error('SImg_RTSIF: File end reached reading header');
    SImg_Free(cc,ri);
    ri.fmt:=TSImg_RawFormat(header.fmt0);
    ri.nx:=header.nx;
    ri.ny:=header.ny;
    ri.nc:=header.nc;
    if not headerOnly then
    begin
        SImg_New(cc,ri,false);
        if ri.dsize<>header.dsize then cc.error('SImg_RTSIF: Invalid size of data');
    end;
    ri.bole:=header.fmt1 and 1<>0;
    ri.cf:=header.cf;
    ri.co:=header.co;
    { read meta data, use image buffer if large enough }
    md:=ri.pt;
    i:=header.ms*sizeof(TSImg_TSIFMeta);
    if (md=nil) or (i>ri.dsize) then getmem2(md,i);
    if (i>0) and (fpread2(fd,pchar(md),i)<i) then cc.error('SImg_RTSIF: File end reached while reading meta data');
    if mi<>nil then
    begin
        for i:=0 to header.ms-1 do
            with md[i] do
            begin
                if (ks>0) and (vs>0) then
                begin
                    cs:=ks+vs;
                    while (cs<126) and (b[cs]<>#0) do
                        inc(cs);
                    cs-=ks+vs;
                    setlength(key,ks);
                    if ks>0 then move(b[0],key[1],ks);
                    setlength(val,vs);
                    if vs>0 then move(b[ks],val[1],vs);
                    setlength(comment,cs);
                    if cs>0 then move(b[ks+vs],comment[1],cs);
                    mi.add(key,val,comment);
                end;
            end;
    end;
    if md<>ri.pt then freemem2(md);
    { read meta data, use image buffer if large enough }
    if not headerOnly then if fpread2(fd, ri.pt, ri.dsize)<>ri.dsize then cc.error('SImg_RTSIF: File end reached while reading image data');
end;

procedure SImg_RTSIF(cc:CSImg_Config; fd:longint; var ri:TSImg_RDImage; mi:CSImg_MetaInfo);
begin
    SImg_RTSIF(cc, fd, ri, mi, false);
end;

{*************************************************************************}
{******* SImg_WTSIF ******************************************************}
{*************************************************************************}
procedure SImg_WTSIF(cc:CSImg_Config; f:CSImg_Output; ri:TSImg_RDImage; mi:CSImg_MetaInfo);
var header              : TSImg_TSIFHeader;
    md                  : ^TSImg_TSIFMeta;
    key,val,comment     : ansistring;
    i,j,cs              : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    cc.debug('SImg_WTSIF: called with: output='+f.info+' ri='+SImg_InfoStr2(ri));
    if not SImg_Recover(cc,ri) then SImg_Error('SImg_WTSIF: Invalid image');
    { build + write header }
    header.nx:=ri.nx;
    header.ny:=ri.ny;
    header.nc:=ri.nc;
    header.dsize:=ri.dsize;
    header.cf:=ri.cf;
    header.co:=ri.co;
    header.fmt0:=ord(ri.fmt);
    header.fmt1:=0;
    if ri.bole then header.fmt1:=header.fmt1 or 1;
    header.ms:=0;
    if mi<>nil then header.ms:=mi.size;
    f.open(cc);
    f.write(cc,@SImg_TSIFHeaderID,4);
    f.write(cc,@header,sizeof(header));
    { write metadata }
    i:=header.ms*sizeof(TSImg_TSIFMeta);
    if i>0 then
    begin
        getmem2(md,i);
        for j:=0 to mi.size-1 do
            with md[j] do
            begin
                key:=mi.key[j];
                val:=mi.val[j];
                if length(key)+length(val)<126 then
                begin
                    comment:=mi.comment[j];
                    ks:=length(key);
                    vs:=length(val);
                    cs:=126-ks-vs;
                    if length(comment)<cs then cs:=length(comment);
                    if ks>0 then move(key[1],b[0],ks);
                    if vs>0 then move(val[1],b[ks],vs);
                    if cs>0 then move(comment[1],b[ks+vs],cs);
                    for cs:=ks+vs+cs to 125 do
                        b[cs]:=#0;
                end
                else begin
                    cc.warning('Key plus value to long for key "'+key+'"');
                    ks:=0;
                    vs:=0;
                    b[0]:=#0;
                end;
            end;
        f.write(cc,md,i);
        freemem2(md,i);
    end;
    { write image data }
    f.write(cc,ri.pt,header.dsize);
    f.close(cc);
end;
