{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{******* file format routines ********************************************}
{*************************************************************************}

{*************************************************************************}
{******* CSImg_Output ****************************************************}
{*************************************************************************}
constructor CSImg_Output.init;
begin
    isopen:=false;
    info:='<undefined>';
end;

procedure CSImg_Output.close(cc:CSImg_Config);
begin
    if isopen then close_int(cc);
    self.destroy;
end;

procedure CSImg_Output.open(cc:CSImg_Config);
begin
    isopen:=true;
    open_int(cc);
end;

procedure CSImg_Output.write(cc:CSImg_Config; pt:pointer; num:int64);
begin
    if cc=nil then cc:=SImg_Config_Default;
    if isopen then write_int(cc,pt,num)
    else cc.error('CSImg_Output.write: Output not open');
end;


{*************************************************************************}
{******* CSImg_FileOutput ************************************************}
{*************************************************************************}
constructor CSImg_FileOutput.init(const theFn:ansistring);
begin
    inherited init;
    fn:=theFn;
    info:='file:'+fn;
end;

procedure CSImg_FileOutput.open_int(cc:CSImg_Config);
begin
    if cc=nil then cc:=SImg_Config_Default;
    if (fn='') or (fn='-') then fd:=1
    else fd:=fpopen(fn,Open_Creat or Open_Trunc or Open_WROnly);
    if fd<0 then cc.error('CSImg_FileOutput.open_int: Can''t open '+fn);
end;

procedure CSImg_FileOutput.close_int(cc:CSImg_Config);
begin
    fpclose(fd);
end;

procedure CSImg_FileOutput.write_int(cc:CSImg_Config; pt:pointer; num:int64);
var i,j : int64;
    p   : pbyte absolute pt;
begin
    i:=0;
    repeat
        j:=fpwrite(fd, p[i], num-i);
        i+=j;
    until (i>=num) or (j<1);
end;


{*************************************************************************}
{********* fpread2 *******************************************************}
{*************************************************************************}
{ tries to read at least len bytes are read
  returns number of bytes read or -1 on error }
function fpread2(fd:longint; buf:pchar; len:aint):aint;
var res : aint;
begin
    result:=0;
    if len<0 then exit;
    res:=1;
    while (res>0) and (len>result) do
    begin
        res:=fpread(fd, buf, len-result);
        if res>0 then
        begin
            buf+=res;
            result+=res;
        end;
    end;
    if res<0 then result:=res;
end;

{*************************************************************************}
{******* SImg_ReadImageFile **********************************************}
{*************************************************************************}
procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var ri:TSImg_RDImage; mi:CSImg_MetaInfo);
const readsize = $100000;
var fd,i,j     : longint;
    d,e        : aint;
    id         : array[0..3] of char;
    pta        : array[0..16*1024] of ^char;    { file size limit is 16 GB }
    fb         : TSimg_FileBuf;
    debuglevel : longint;
    finfo      : stat;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    debuglevel:=SImg_DebugLevel;
    cc.debug('SImg_ReadImageFile: called with: file='+fn+' ri='+SImg_InfoStr2(ri));

    if (fn='') or (fn='-') then
    begin
        if IsATTY(stdinputhandle)>0 then cc.Error('SImg_ReadImageFile: Image input via terminal');
        fn:='/dev/stdin';
    end;

    if not FileExists(fn) then cc.Error('SImg_ReadImageFile: File not found: '+fn);
    fd:=fpopen(fn,o_rdOnly);
    if fd<0 then cc.error('CSImg_ReadImageFile: '+fn+': Error opening file: '+StrError(fpGetErrno));
    if fpread2(fd, @id, 4)<4 then cc.error('CSImg_ReadImageFile: '+fn+': File empty');

    if (id=SImg_TSIFHeaderID) then
    begin
        SImg_RTSIF(cc, fd, ri, mi);
        fpclose(fd);
        SImg_DebugLevel:=debuglevel;
        SImg_DecDebugLevel;
        exit;
    end;

    e:=0;
    if fpfstat(fd,finfo)=0 then e:=finfo.st_size;
    if e<=readsize then         // slow read with copying buffers if size is unknown (i.e. via terminal)
    begin
        getmem2(pta[0],readsize);
        move(id,pta[0]^,4);
        j:=fpread2(fd,pta[0]+4,readsize-4);
        if j>=0 then j+=4;
        d:=0;
        while j=readsize do
        begin
            d+=1;
            getmem2(pta[d],readsize);
            j:=fpread2(fd,pta[d],readsize);
        end;
        fb.size:=d*readsize + j;
        if (j<0) or (fb.size<256) then
        begin
            for i:=0 to d do
                freemem2(pta[i],readsize);
            if j<0 then cc.error('CSImg_ReadImageFile: '+fn+': Error reading file: '+StrError(fpGetErrno))
            else cc.error('SImg_ReadImageFile: File to small: '+fn);
        end;
        fpclose(fd);
//      writeln('size=',fb.size,'  d=',d);
        if d=0 then
        begin
            fb.reserv:=readsize;
            fb.pt:=pbyte(pta[0]);
        end
        else begin
            fb.reserv:=d*readsize + j;
            getmem2(fb.pt, fb.reserv);
            e:=0;
            for i:=0 to d-1 do
            begin
                move( pta[i]^, fb.pt[e], readsize);
                e:=e+readsize;
            end;
            if j>0 then move(pta[d]^,fb.pt[e],j);
            for i:=0 to d do
                freemem2(pta[i],readsize);
        end;
    end
    else begin  // direct read into buffer
        fb.size:=e;
        fb.reserv:=e;
        getmem2(fb.pt, fb.reserv);
        move(id,fb.pt^,4);
        e:=fpread2(fd, pchar(fb.pt+4), fb.size-4);
        if (e<252) then
        begin
            freemem2(fb.pt, fb.reserv);
            if e<0 then cc.error('CSImg_ReadImageFile: '+fn+': Error reading file: '+StrError(fpGetErrno))
            else cc.error('SImg_ReadImageFile: File to small: '+fn);
        end;
        fb.size:=e+4;
//      writeln('reserved=',fb.reserv,' size=',fb.size);
    end;

    SImg_Free(cc,ri);
    if not SImg_RFITS(cc,fb,ri,mi) then
    begin
        SImg_DebugLevel:=debuglevel;
        cc.debug('SImg_ReadImagefile(2): FITS failed ...');
        if not SImg_RSIF(cc,fb,ri,mi) then
        begin
            SImg_DebugLevel:=debuglevel;
            cc.debug('SImg_ReadImagefile(3): SIF failed ...');
            if not SImg_RPNM(cc,fb,ri,mi) then
            begin
                SImg_DebugLevel:=debuglevel;
                cc.debug('SImg_ReadImagefile(4): PNM failed ...');
                if not SImg_RSBIG3(cc,fb,ri,mi) then
                begin
                    SImg_DebugLevel:=debuglevel;
                    cc.debug('SImg_ReadImagefile(5): SBIG3 failed ...');
                end;
            end;
        end;
    end;
    freemem2(fb.pt,fb.reserv);
    if not SImg_Recover(cc,ri,mi) then cc.error('SImg_ReadImageFile: Image format not supported');
    SImg_DebugLevel:=debuglevel;
    SImg_DecDebugLevel;
end;

procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var fi:TSImg_FPImage; mi:CSImg_MetaInfo);
var ri  : TSImg_RDImage;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_ReadImageFile(cc,fn,ri,mi);
    SImg_Convert(cc,ri,fi,mi,true,false);
end;

procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var ri:TSImg_RDImage);
begin
    SImg_ReadImageFile(cc, fn, ri, nil);
end;

procedure SImg_ReadImageFile(cc:CSImg_Config; fn:ansistring; var fi:TSImg_FPImage);
begin
    SImg_ReadImageFile(cc, fn, fi, nil);
end;


{*************************************************************************}
{******* SImg_RHeader ****************************************************}
{*************************************************************************}
{ ri only contains image size }
procedure SImg_RHeader(cc:CSImg_Config; fn:ansistring; var ri:TSImg_RDImage; mi:CSImg_MetaInfo);
const maxSize = 2880*128;  // up to 4608 cards
var fb          : TSImg_FileBuf;
    fd,j        : longint;
    debuglevel  : longint;
    id          : array[0..3] of char;
    b           : boolean;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    debuglevel:=SImg_DebugLevel;
    cc.debug('SImg_RHeader: called with: file='+fn+' ri='+SImg_InfoStr2(ri));
    if (fn='') or (fn='-') then
    begin
        if IsATTY(stdinputhandle)>0 then cc.Error('SImg_RHeader: Image input via terminal');
        fn:='/dev/stdin';
    end;
    if not FileExists(fn) then cc.Error('SImg_RHeader: File not found: '+fn);
    fd:=fpopen(fn,o_rdOnly);
    if fd<0 then cc.error('CSImg_RHeader: '+fn+': Error opening file: '+StrError(fpGetErrno));

    if fpread2(fd, @id, 4)<4 then cc.error('CSImg_ReadHeader: '+fn+': File empty');

    { TSIF }
    if (id=SImg_TSIFHeaderID) then
    begin
        SImg_RTSIF(cc, fd, ri, mi, true);
        fpclose(fd);
        SImg_DebugLevel:=debuglevel;
        SImg_DecDebugLevel;
        exit;
    end;

    { FITS }
    getmem2(fb.pt,maxSize);
    fb.reserv:=maxSize;
    move(id,fb.pt^,4);
    j:=fpread2(fd, pchar(fb.pt+4), maxSize-4)+4;
    fb.size:=j;
    b:=(j>=2880) and SImg_RFITS(cc,fb,ri,mi,true);
    freemem2(fb.pt,fb.reserv);
    fpclose(fd);
    if j<0 then cc.error('CSImg_RHeader: '+fn+': Error reading file: '+StrError(fpGetErrno))
    else if j<2880 then cc.error('SImg_RHeader: File to small or empty: '+fn)
    else if not b then cc.error('SImg_ReadHeader: Not a FITS file: '+fn);
    SImg_DebugLevel:=debuglevel;
    SImg_DecDebugLevel;
end;


procedure SImg_RHeader(cc:CSImg_Config; fn:ansistring; mi:CSImg_MetaInfo);
var ri   : TSImg_RDImage;
begin
    ri.pt:=nil;
    SImg_RHeader(cc,fn,ri,mi);
end;


{*************************************************************************}
{******* SImg_WriteImageFile *********************************************}
{*************************************************************************}
procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; ri:TSImg_RDImage; mi:CSImg_MetaInfo);
var s  : string[8];
    i  : longint;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_WriteImageFile: called with: file='+fn+' ri='+SImg_InfoStr2(ri));
    s:='';
    i:=length(fn);
    while (i>0) and (fn[i]<>'.') do
        dec(i);
    s:=copy(fn,i,length(fn)-i+1);
    upstr(s);
    if (s='') or (s[1]<>'.') or (s='.TSIF') then SImg_WTSIF(cc,CSImg_FileOutput.init(fn),ri, mi)
    else if (length(s)=4) and (s[2]='P') and (s[4]='M') then SImg_WPNM(cc,CSImg_FileOutput.init(fn),ri, mi)
    else if s='.SIF' then SImg_WSIF(cc,CSImg_FileOutput.init(fn),ri, mi)
    else SImg_WFITS(cc,CSImg_FileOutput.init(fn),ri,mi).close(cc);
    SImg_DecDebugLevel;
end;

procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; fi:TSImg_FPImage; mi:CSImg_MetaInfo);
var ri : TSImg_RDImage;
begin
    if cc=nil then cc:=SImg_Config_Default;
    SImg_IncDebugLevel;
    cc.debug('SImg_WriteImageFile: called with: file='+fn+' fi='+SImg_InfoStr2(fi));
    ri.pt:=nil; SImg_Convert(cc,fi,ri,true,true);
    SImg_WriteImageFile(cc,fn,ri,mi);
    if ri.pt<>fi.pt then
    begin
        SImg_Free(cc,ri);
        ri.pt:=nil;
    end;
    SImg_DecDebugLevel;
end;

procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; ri:TSImg_RDImage);
begin
    SImg_WriteImageFile(cc, fn, ri, nil);
end;

procedure SImg_WriteImageFile(cc:CSImg_Config; const fn:ansistring; fi:TSImg_FPImage);
begin
    SImg_WriteImageFile(cc, fn, fi, nil);
end;

