{  simg -- image processing library
   Version 20220317
   Copyright (C) 1999-2022 Stefan Ziegenbalg
  
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License version 3 as
   published by the Free Software Foundation.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, see http://www.gnu.org/licenses/.
!}

{*************************************************************************}
{*************************************************************************}
{******* system routines for FPImages ************************************}
{*************************************************************************}
{*************************************************************************}


{*************************************************************************}
{******* SImg_GetImageTableEntry *****************************************}
{*************************************************************************}
function SImg_GetImageTableEntry(fi:TSImg_FPImage):longint;
begin
    result:=SImg_GetImageTableEntry(fi.pt, fi.size*sizeof(simgfp));
end;


{*************************************************************************}
{******* SImg_New ********************************************************}
{*************************************************************************}
{ spt: nil: init with zeros if clear=true }
procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage; clear:boolean);  { nx, ny and nc must be set }
var i  : longint;
    d  : ^simgfp;
begin
    if cc=nil then cc:=SImg_Config_default;
    SImg_Free(cc,r);
    if clear then SImg_IncDebugLevel;
    if (r.nx<minxsize) or (r.nx>maxxsize) or (r.ny<minysize) or (r.ny>maxysize) or (r.nc<mincnsize) or (r.nc>maxcnsize) then cc.error('SImg_New: illegal dimensions');
    r.sl:=r.nx*r.nc;
    r.size:=r.sl*r.ny;
    i:=SImg_GetEmptyImageTableEntry;
    SImg_Sem_Lock(SImg_Sem_Img);
    try
        with SImg_ImageTable[i] do
        begin
            size:=r.size*sizeof(simgfp);
            getmem2(pt,size);
            r.pt:=pt;
        end;
    finally
            SImg_Sem_Unlock(SImg_Sem_Img);
    end;
    r.iid:=i;
    cc.debug('SImg_New: allocated '+base1024str(r.size*sizeof(simgfp))+' bytes for '+SImg_InfoStr2(r)+' at '+ptr2str(r.pt));
    if not SImg_Integrity(r) then cc.error('SImg_New: internal error');
    if clear then
    begin
        d:=r.pt;
        for i:=0 to r.size-1 do
            d[i]:=0;
    end;
    if clear then SImg_DecDebugLevel;
end;

procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage; nx,ny,nc:longint; clear:boolean);
begin
    r.nx:=nx; r.ny:=ny; r.nc:=nc;
    SImg_New(cc,r, clear);
end;

procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage);
begin
    SImg_New(cc,r, true);
end;

procedure SImg_New(cc:CSImg_Config; var r:TSImg_FPImage; nx,ny,nc:longint);
begin
    r.nx:=nx; r.ny:=ny; r.nc:=nc;
    SImg_New(cc,r, true);
end;


{*************************************************************************}
{******* SImg_Free *******************************************************}
{*************************************************************************}
procedure SImg_Free(cc:CSImg_Config; var fi:TSImg_FPImage);
var i  : longint;
begin
    if cc=nil then cc:=SImg_Config_default;
    i:=SImg_GetImageTableEntry(fi);
    if (i>=0) then
        with SImg_ImageTable[i] do
        begin
            cc.debug('SImg_Free: deallocating '+base1024str(size)+'bytes for '+SImg_InfoStr2(fi)+' at '+ptr2str(pt));
            SImg_Sem_Lock(SImg_Sem_Img);
            try
                freemem2(pt,size);
                pt:=nil; size:=0;
            finally
                SImg_Sem_Unlock(SImg_Sem_Img);
            end;
        end;
    fi.size:=0; fi.pt:=nil; fi.iid:=-1;
end;


{*************************************************************************}
{******* SImg_Move *******************************************************}
{*************************************************************************}
procedure SImg_Move(cc:CSImg_Config; var fi,fj:TSImg_FPImage);  { fi -> fj }
begin
    if not SImg_Overwritealert(fi,fj) then
    begin
        SImg_Free(cc,fj);
        fj:=fi;
        fi.pt:=nil; fi.size:=0; fi.iid:=-1;
    end;
end;


{*************************************************************************}
{******* SImg_CheckConstant **********************************************}
{*************************************************************************}
function SImg_CheckConstant(fi:TSImg_FPImage):boolean;
var i     : longint;
    d     : ^simgfp;
    r,s,t : simgfp;
begin
    d:=fi.pt;
    r:=d^; s:=r;
    SImg_CheckConstant:=false;
    for i:=0 to fi.size-1 do
    begin
        t:=d^; d+=1;
        if t<r then r:=t;
        if t>s then s:=t;
        if (i and 1023=0) and (abs(r-s)>=simg_mina) then exit;
    end;
    SImg_CheckConstant:=abs(r-s)<simg_mina;
end;

{*************************************************************************}
{******* SImg_Integrity **************************************************}
{*************************************************************************}
function SImg_Integrity(fi:TSImg_FPImage):boolean;
begin
    SImg_Integrity:=(fi.iid>=0) and (fi.iid<SImg_ImageTableSize) and
        (SImg_GetImageTableEntry(fi)=fi.iid) and
        (fi.nx>=minxsize) and (fi.nx<=maxxsize) and
        (fi.ny>=minysize) and (fi.ny<=maxysize) and
        (fi.nc>=mincnsize) and (fi.nc<=maxcnsize) and
        (fi.sl>=fi.nx*fi.nc) and (fi.size>=fi.ny*fi.sl);
end;


{*************************************************************************}
{******* SImg_Recover ****************************************************}
{*************************************************************************}
function SImg_Recover(cc:CSImg_Config;var fi:TSImg_FPImage):boolean;  { true, if fi valid }
var  i   : longint;
begin
    if cc=nil then cc:=SImg_Config_default;
    i:=SImg_GetImageTableEntry(fi);
    SImg_Recover:=(i>=0) and
        (fi.nx>=minxsize) and (fi.nx<=maxxsize) and
        (fi.ny>=minysize) and (fi.ny<=maxysize) and
        (fi.nc>=mincnsize) and (fi.nc<=maxcnsize);
    if SImg_Recover then
    begin
        if fi.iid<>i then
        begin
            cc.warning('SImg_Recover: invalid iid: corrected');
            fi.iid:=i;
        end;
        if fi.sl<fi.nx*fi.nc then
        begin
            cc.warning('SImg_Recover: invalid sl: corrected');
            fi.sl:=fi.nc*fi.nc;
        end;
        if fi.size<fi.sl*fi.ny then
        begin
            cc.warning('SImg_Recover: invalid size: corrected');
            fi.size:=fi.sl*fi.ny;
        end;
    end;
end;


{*************************************************************************}
{******* SImg_OverwriteAlert *********************************************}
{*************************************************************************}
function SImg_OverwriteAlert(fi,fj:TSImg_FPImage):boolean;
var i,j:longint;
begin
    i:=SImg_GetImageTableEntry(fi);
    j:=SImg_GetImageTableEntry(fj);
    SImg_OverwriteAlert:=(i=j) and (i>=0);
end;


{*************************************************************************}
{******* SImg_Duplicate **************************************************}
{*************************************************************************}
procedure SImg_Duplicate(cc:CSImg_Config; fi:TSImg_FPImage; var fj:TSImg_FPImage; dat:boolean);   { fi --> fj; dat=true: copy data }
begin
    if cc=nil then cc:=SImg_Config_default;
    if dat then
    begin
        SImg_IncDebugLevel;
        cc.debug('SImg_Duplicate: Called with: fi='+SImg_InfoStr2(fi)+' fj='+SImg_InfoStr2(fj));
    end;
    if not SImg_Recover(cc,fi) then cc.error('SImg_Duplicate: illegal image: fi');
    if SImg_Overwritealert(fi,fj) then
    begin
        if not dat then SImg_Clear(cc,fj);
        exit;
    end;
    fj.nx:=fi.nx;
    fj.ny:=fi.ny;
    fj.nc:=fi.nc;
    SImg_New(cc,fj,not dat);
    if dat then move(fi.pt^, fj.pt^, fi.size*sizeof(simgfp));
    if dat then SImg_DecDebugLevel;
end;


{*************************************************************************}
{******* SImg_InfoStr ****************************************************}
{*************************************************************************}
function SImg_InfoStr(fi:TSImg_FPImage):string19;
begin
    if SImg_GetImageTableEntry(fi)>=0
    then SImg_InfoStr:=int2str(fi.nx)+'x'+int2str(fi.ny)+'x'+int2str(fi.nc)
    else SImg_InfoStr:='(unused)';
end;

function SImg_InfoStr2(fi:TSImg_FPImage):string37;
begin
    SImg_InfoStr2:='FPImg(id='+int2str(SImg_GetImageTableEntry(fi))+', fmt='+SImg_InfoStr(fi)+')';
end;

